import astropy.units as u
import numpy as np
from astropy.io.fits import getdata
from astropy.time import Time
from scipy.interpolate import RectBivariateSpline

import exoverses.base as base


class ExovistaPlanet(base.planet.Planet):
    """
    Class for the planets in the exoVista systems
    Assumes e=0, which is the case for all downloaded systems
    """

    def __init__(self, infile, fits_ext, star):
        # Get the object's data from the fits file
        with open(infile, "rb") as f:
            obj_data, obj_header = getdata(f, ext=fits_ext, header=True, memmap=False)
            self._wavelengths = getdata(f, ext=0, header=False, memmap=False) * u.um

        self.star = star
        # Time data, setting default epoch to the year 2000
        self._t = Time(2000 + obj_data[:, 0], format="decimalyear")
        self.t0 = self._t[0]

        # self._x_pix = obj_data[:, 1] * u.pixel
        # self._y_pix = obj_data[:, 2] * u.pixel
        # _x_pix_interp = interp1d(self._t.jd, self._x_pix.value, kind="cubic")
        # _y_pix_interp = interp1d(self._t.jd, self._y_pix.value, kind="cubic")
        # self._x_pix_interp = lambda t: _x_pix_interp(t.jd)
        # self._y_pix_interp = lambda t: _y_pix_interp(t.jd)

        # Barycentric position data
        self._x = obj_data[:, 9] * u.AU
        self._y = obj_data[:, 10] * u.AU
        self._z = obj_data[:, 11] * u.AU

        # Barycentric velocity data
        self._vx = obj_data[:, 12] * u.AU / u.yr
        self._vy = obj_data[:, 13] * u.AU / u.yr
        self._vz = obj_data[:, 14] * u.AU / u.yr

        # Assign the planet's time-varying mean anomaly, argument of pericenter,
        # true anomaly, and contrast
        self.rep_w = (obj_data[:, 7] * u.deg) % (2 * np.pi * u.rad)
        self.M = (obj_data[:, 8] * u.deg) % (2 * np.pi * u.rad)
        # true anomaly for circular orbits
        self.nu = (self.rep_w + self.M) % (2 * np.pi * u.rad)
        self.phase_angles = obj_data[:, 15]

        self.contrast = obj_data[:, 16:]
        self.cont_interp = RectBivariateSpline(
            self._wavelengths,
            self._t.decimalyear * u.yr,
            self.contrast.T,
            kx=4,
            ky=4,
        )

        # Spectral flux density of the planet
        self.planet_spec_flux_density_interp = RectBivariateSpline(
            self._wavelengths.to_value(u.nm),
            self._t.decimalyear,
            np.multiply(self.contrast, star._star_flux_density).T,
            kx=4,
            ky=4,
        )

        # Initial mean anomaly
        self.M0 = self.M[0]
        planet_dict = {
            "t0": self.t0,
            "a": obj_header["A"] * u.AU,
            "e": obj_header["E"],
            "inc": (obj_header["I"] * u.deg),
            "W": (obj_header["LONGNODE"] * u.deg),
            "w": (obj_header["ARGPERI"] * u.deg),
            "mass": obj_header["M"] * u.M_earth,
            "radius": obj_header["R"] * u.R_earth,
            "M0": self.M0,
            "p": 0.2,
        }
        base.planet.Planet.__init__(self, planet_dict, star)
        self.solve_dependent_params()

        self.star = star

        self.classify_planet()

    def spec_flux_density(self, wavelengths, times):
        """
        Calculate the spectral flux density of the star at the given wavelengths
        and times
        Args:
            wavelengths (astropy Quantity array):
                Wavelengths to calculate spectral flux density
            times (astropy Time array):
                Times to calculate spectral flux density
        Returns:
            F (astropy Quantity array):
                Spectral flux density values
        """
        return (
            self.planet_spec_flux_density_interp(
                wavelengths.to_value(u.nm), times.decimalyear
            ).T
            << u.Jy
        )

    # def rotate_to_sky_coords(self, vec, roll_angle=0 * u.rad):
    #     """
    #     Rotate from barycentric coordinates to plane of the sky, this is set up
    #     to match the exovista data

    #     Args:
    #         vec (np.array):
    #             Nx3 array of vectors in system-plane coordinates
    #         roll_angle (astropy Quantity):
    #             Angle to rotate the vectors by in the plane of the sky to simulate
    #             roll of the telescope

    #     Returns:
    #         vec (np.array):
    #             Nx3 array of vectors rotated to sky coordinates

    #     """
    #     # Rotate around x axis with midplane inclination
    #     vec = misc.rotate_vectors(vec.T, [1, 0, 0], -self.star.midplane_I)

    #     # Rotate around z axis with midplane position angle
    #     vec = misc.rotate_vectors(vec, [0, 0, 1], self.star.midplane_PA)

    #     # Flip around z axis
    #     vec[:, 2] = -vec[:, 2]

    #     if roll_angle != 0 * u.rad:
    #         # Rotate around y axis with roll angle
    #         vec = misc.rotate_vectors(vec, [0, 0, 1], roll_angle)
    #     return vec



import astropy.units as u
import pandas as pd

from exoverses.base.planet import Planet


class ExosimsPlanet(Planet):
    """
    Class for the planets in the EXOSIMS systems
    Assumes e=0, which is the case for all downloaded systems
    """

    def __init__(self, SU, star, pInd, t0):
        self.star = star

        # Time data
        self._t = t0.decimalyear * u.yr
        self.t0 = t0

        # Position data
        self._x = SU.r[pInd][0]
        self._y = SU.r[pInd][1]
        self._z = SU.r[pInd][2]

        # Velocity data
        self._vx = SU.v[pInd][0]
        self._vy = SU.v[pInd][1]
        self._vz = SU.v[pInd][2]

        # Assign the planet's keplerian orbital elements
        # self.a = SU.a[pInd]
        # self.e = SU.e[pInd]
        # self.inc = SU.I[pInd]
        # self.W = SU.O[pInd]
        # self.w = SU.w[pInd]

        # # Assign the planet's mass/radius information
        # self.mass = SU.Mp[pInd]
        # self.radius = SU.Rp[pInd]
        # # Initial mean anomaly
        # self.M0 = SU.M0[pInd]

        planet_dict = {
            "t0": t0,
            "a": SU.a[pInd],
            "e": SU.e[pInd],
            "inc": SU.I[pInd],
            "W": SU.O[pInd],
            "w": SU.w[pInd],
            "mass": SU.Mp[pInd],
            "radius": SU.Rp[pInd],
            "M0": SU.M0[pInd],
            "p": 0.2,
        }
        Planet.__init__(self, planet_dict)
        self.solve_dependent_params()
        # Propagation table
        self.vectors = pd.DataFrame(
            {
                "t": [self._t.decompose().value],
                "x": [self._x.decompose().value],
                "y": [self._y.decompose().value],
                "z": [self._z.decompose().value],
                "vx": [self._vx.decompose().value],
                "vy": [self._vy.decompose().value],
                "vz": [self._vz.decompose().value],
            }
        )



import astropy.constants as const
import astropy.units as u
import numpy as np
from astropy.time import Time

import exoverses.util.misc as misc
from exoverses.base.planet import Planet


class FitPlanet(Planet):
    def __init__(self, planet_dict, true_system):
        """
        filling in parameters that are not found in the fitting process.
        Assumes that the planet_dict includes
        T, Tc, secosw, sesinw, K
        basis.
        """
        self.star = true_system.star
        for att, value in planet_dict.items():
            setattr(self, att, value)
        self.e = self.secosw**2 + self.sesinw**2
        self.w_s = np.arctan2(self.sesinw, self.secosw) * u.rad
        self.w = (self.w_s + np.pi * u.rad) % (2 * np.pi * u.rad)
        self.w_p = self.w
        self.T_p = misc.timetrans_to_timeperi(
            self.T_c.jd, self.T.to(u.d).value, self.e, self.w_s.to(u.rad).value
        )
        self.msini = (
            misc.Msini(
                self.K.decompose().value,
                self.T.to(u.d).value,
                self.star.mass.to(u.M_sun).value,
                self.e,
            )
            * u.M_earth
        )
        self.mu = (const.G * self.star.mass).decompose()
        self.a = ((self.mu * (self.T / (2 * np.pi)) ** 2) ** (1 / 3)).decompose()

        # Finding the mean anomaly at time of conjunction
        nu_p = (np.pi / 2 * u.rad - self.w_s) % (2 * np.pi * u.rad)
        E_p = 2 * np.arctan2(
            np.sqrt((1 - self.e)) * np.tan(nu_p / 2), np.sqrt((1 + self.e))
        )
        self.M0 = (E_p - self.e * np.sin(E_p) * u.rad) % (2 * np.pi * u.rad)
        self.t0 = Time(self.T_c.jd, format="jd")

        self.compare_to_system(true_system)

    def compare_to_system(self, true_system):
        """
        Function to check how well a planet fit compares to the true system.
        Gets the index of the planet that the fit is most likely describing
        and the RMS error of the fit.
        """
        # Not doing argument of periastron because it is finicky for circular
        # orbits
        # werr = true_system.getpattr('w') - fitted_planet.w
        eerr = true_system.getpattr("e") - self.e
        Kerr = true_system.getpattr("K") - self.K
        Terr = true_system.getpattr("T") - self.T

        # Normalize the differences by the range of the system values
        Knorm = Kerr / (np.ptp(true_system.getpattr("K")))
        Tnorm = Terr / (np.ptp(true_system.getpattr("T")))

        self.all_rms = np.sqrt(eerr**2 + Knorm**2 + Tnorm**2)
        self.best_match = np.argmin(self.all_rms)
        self.best_rms = self.all_rms[self.best_match]



import copy

import astropy.units as u
import numpy as np
import pandas as pd
from astropy.time import Time

from exoverses.base.system import System
from exoverses.fit.planet import FitPlanet


class FitSystem(System):
    def __init__(self, search_obj, true_system):
        """
        This function creates a rvtools system object with the planets found
        during the fitting process. It also calculates the planet that the
        fit is most likely describing.
        """
        search_params = search_obj.post.params
        self.star = copy.deepcopy(true_system.star)
        self.planets = []
        for nplan in range(1, search_obj.num_planets + 1):
            # Create dictionary to initialize planet with
            plan_dict = {}
            plan_dict["T"] = search_params[f"per{nplan}"].value * u.d
            plan_dict["T_c"] = Time(search_params[f"tc{nplan}"].value, format="jd")
            plan_dict["secosw"] = search_params[f"secosw{nplan}"].value
            plan_dict["sesinw"] = search_params[f"sesinw{nplan}"].value
            plan_dict["K"] = search_params[f"k{nplan}"].value * u.m / u.s

            # Initialize the planet object
            planet = FitPlanet(plan_dict, true_system)
            self.planets.append(planet)
        self.pInds = np.arange(0, search_obj.num_planets)
        self.cleanup()
        self.true_system = true_system

    def get_p_df(self):
        patts = [
            "K",
            "T",
            "secosw",
            "sesinw",
            "T_c",
            "a",
            "e",
            "w",
            "M0",
            "t0",
            "msini",
            "best_match",
            "best_rms",
        ]
        p_df = pd.DataFrame()
        for att in patts:
            pattr = self.getpattr(att)
            if isinstance(pattr, u.Quantity):
                p_df[att] = pattr.value
            else:
                p_df[att] = pattr

        return p_df



import astropy.units as u
import numpy as np
from astropy.io.fits import getheader
from keplertools import fun as kt

import exoverses.exovista as ev
from exoverses.base.system import System


class ExovistaSystem(System):
    """
    Class for the whole stellar system

    Args:
        infile (Path):
            Path to the exoVista fits file
    """

    def __init__(self, infile, initial_epoc=2000, convert=False, filter=False):
        self.file = infile
        self.file_name = self.file.name

        # fits file extensions, exoVista hard codes these
        planet_ext = 5
        disk_ext = 2

        # Get the number of planets
        with open(infile, "rb") as f:
            # read header of first extension
            h = getheader(f, ext=0, memmap=False)
        # get the largest extension
        n_ext = h["N_EXT"]
        nplanets = n_ext - 4

        # Create star object
        self.star = ev.star.ExovistaStar(infile)
        self.nstar_wavelengths = len(self.star._wavelengths)
        self.nstar_times = len(self.star._t)
        self.planets = []
        # loop over all planets
        for i in range(nplanets):
            self.planets.append(
                ev.planet.ExovistaPlanet(infile, planet_ext + i, self.star)
            )
        self.disk = ev.disk.ExovistaDisk(infile, disk_ext, self.star)
        self.ndisk_wavelengths = len(self.disk._wavelengths)

        # mas/pixel
        self.pixel_scale = self.star.pixel_scale

        # What the nbody simulation frame is
        self.nbody_frame = "bary-sky"
        self.origin = "ExoVista"

        self.planet_cleanup()
        self.star_cleanup()

        if convert:
            # Use nbody integration to get the positions in the sky-frame
            _t0 = self.planets[0]._t[0].reshape(1)
            bary = self.propagate(_t0, clean=True, ref_frame="bary-sky", prop="nbody")

            # Now we're converting the system object to the "bary" frame from
            # the "bary-sky" frame without a midplane
            self.nbody_frame = "bary"

            # Set the star's position and velocity
            sel = bary.sel(object="star", index=0, ref_frame="bary-sky").squeeze()
            _x, _y, _z = sel.x.item(), sel.y.item(), sel.z.item()
            _vx, _vy, _vz = sel.vx.item(), sel.vy.item(), sel.vz.item()
            self.star._x = u.Quantity(_x * u.m).reshape(1)
            self.star._y = u.Quantity(_y * u.m).reshape(1)
            self.star._z = u.Quantity(_z * u.m).reshape(1)
            self.star._vx = u.Quantity(_vx * u.m / u.s).reshape(1)
            self.star._vy = u.Quantity(_vy * u.m / u.s).reshape(1)
            self.star._vz = u.Quantity(_vz * u.m / u.s).reshape(1)

            # Set the planet's positions, velocities, and orbital angles
            for i, planet in enumerate(self.planets):
                sel = bary.sel(object="planet", index=i, ref_frame="bary-sky").squeeze()
                _x, _y, _z = sel.x.item(), sel.y.item(), sel.z.item()
                _vx, _vy, _vz = sel.vx.item(), sel.vy.item(), sel.vz.item()
                # Assign the new positions and velocities
                planet._x = u.Quantity(_x * u.m).reshape(1)
                planet._y = u.Quantity(_y * u.m).reshape(1)
                planet._z = u.Quantity(_z * u.m).reshape(1)
                planet._vx = u.Quantity(_vx * u.m / u.s).reshape(1)
                planet._vy = u.Quantity(_vy * u.m / u.s).reshape(1)
                planet._vz = u.Quantity(_vz * u.m / u.s).reshape(1)

                # Calculate the orbital elements
                rs = np.array([_x, _y, _z])
                vs = np.array([_vx, _vy, _vz])
                mu = planet.mu.decompose().value.reshape(1)
                _, _, _E, _O, _I, _w, _P, _tau = kt.vec2orbElem(rs, vs, mu)

                # Set the angles
                planet.W = (_O[0] * u.rad).to(u.deg)
                planet.w = (_w[0] * u.rad).to(u.deg)
                planet.inc = (_I[0] * u.rad).to(u.deg)
                _E = _E[0] * u.rad
                planet.M0 = (
                    (_E.value - planet.e * np.sin(_E)) % (2 * np.pi) * u.rad
                ).to(u.deg)
                planet.solve_dependent_params()
                planet.classify_planet()
        if filter:
            # TODO: Make this more flexible
            # Keep only the planets that are Earth-like
            earth_ind = self.getpattr("is_earth")
            # K_vals = self.getpattr("K")
            # max_K = K_vals.max()
            # max_earth_K = self.getpattr("K")[np.argwhere(earth_ind).ravel()]
            # Remove all planets with K values larger than the largest K value of the Earth
            # except the larget K value
            # to_remove = np.argwhere(
            #     (K_vals > max_earth_K.max()) & (K_vals != max_K)
            # ).ravel()
            self.planets = [
                self.planets[i] for i in range(len(self.planets)) if earth_ind[i]
            ]
            self.planet_cleanup()

    def spec_flux_densities(self, wavelengths, times):
        """
        Calculate the spectral flux density of the system at the given times and
        wavelengths
        Args:
            wavelengths (astropy Quantity array):
                Wavelengths to calculate the spectral flux density at
            times (astropy Time array):
                Times to calculate the spectral flux density at
        Returns:
            star_flux_density (astropy Quantity array):
                Spectral flux density of the star at the given times in Jy
                The shape of the cube is (Ntimes, Nwavelengths)
            planet_flux_density (astropy Quantity array):
                Spectral flux density of the planets at the given times in Jy.
                The shape of the cube is (Nplanets, Ntimes, Nwavelengths)
            disk_flux_density (astropy Quantity array):
                Spectral flux density of the disk at the given times in Jy per
                pixel. The shape of the cube is (Ntimes, Nwavelengths, Npix,
                                                 Npix)

        """
        # Calculate the spectral flux density of the star
        star_flux_density = self.star.spec_flux_density(wavelengths, times)

        # Calculate the spectral flux density of the planets
        planet_flux_density = (
            np.zeros((len(self.planets), len(times), len(wavelengths))) * u.Jy
        )
        for i, planet in enumerate(self.planets):
            planet_flux_density[i] += planet.spec_flux_density(wavelengths, times)

        # Calculate the spectral flux density of the disk
        disk_flux_density = self.disk.spec_flux_density(wavelengths, times)
        return star_flux_density, planet_flux_density, disk_flux_density

    def add_objects_to_rebound(self, sim):
        """
        Method that adds the star and planets to the rebound simulation
        """
        sim.add(
            m=self.star.mass.decompose().value,
            x=self.star._x[0].decompose().value,
            y=self.star._y[0].decompose().value,
            z=self.star._z[0].decompose().value,
            vx=self.star._vx[0].decompose().value,
            vy=self.star._vy[0].decompose().value,
            vz=self.star._vz[0].decompose().value,
        )
        for planet in self.planets:
            sim.add(
                m=planet.mass.decompose().value,
                x=planet._x[0].decompose().value,
                y=planet._y[0].decompose().value,
                z=planet._z[0].decompose().value,
                vx=planet._vx[0].decompose().value,
                vy=planet._vy[0].decompose().value,
                vz=planet._vz[0].decompose().value,
            )
        return sim



import multiprocessing
import os
import subprocess
from pathlib import Path

import astropy.io.fits as fits
import dill
import numpy as np
import pandas as pd
from ExoVista import (
    Settings,
    generate_disks,
    generate_planets,
    generate_scene,
    load_stars,
    read_solarsystem,
)
from tqdm import tqdm

from exoverses.base.universe import Universe
from exoverses.exovista.system import ExovistaSystem


def create_universe(universe_params, workers=10):
    data_path = Path(universe_params["data_path"])
    targetlist = universe_params["target_list"]
    convert = universe_params.get("convert")
    filter = universe_params.get("filter", False)
    seed = universe_params["forced_seed"]
    # system_path is used to generate a specific system
    system_path = universe_params.get("system_path")
    full_path = f"{data_path}"
    has_system_path = system_path is not None
    if has_system_path:
        settings = Settings.Settings(timemax=10.0, ncomponents=2, output_dir=full_path)
        s, p, a, d, c, new_settings = read_solarsystem.read_solarsystem(
            settings, system_file=system_path
        )
        generate_scene.generate_scene(s, p, d, a, c, new_settings)
    else:
        generate_systems(targetlist, full_path, seed, workers=workers)

    universe = ExovistaUniverse(
        full_path, targetlist, convert=convert, cache=True, filter=filter
    )
    return universe


class ExovistaUniverse(Universe):
    """
    Class for the whole exoVista universe
    """

    def __init__(
        self,
        path,
        target_list,
        convert=False,
        cache=False,
        initial_epoch=2000,
        filter=False,
    ):
        """
        Args:
            path (str or Path):
                Location of all the system files. Should be something like "data/1/"
        """
        self.type = "ExoVista"
        if cache:
            cache_base = Path(*(Path(path).parts[:-1]))
            if not cache_base.exists():
                cache_base.mkdir(parents=True)
        self.path = path

        # Load all systems
        p = Path(path).glob("*.fits")
        system_files = [x for x in p if x.is_file]
        hip_inds = pd.read_csv(target_list)
        if "HIP" not in hip_inds.columns:
            hip_inds = hip_inds.rename(columns={"name": "HIP"})
        hip_inds["HIP"] = hip_inds["HIP"].str.replace("HIP ", "").astype(int)
        relevant_files = []
        for file in system_files:
            file_hip_ind = int(file.stem.split("-")[1].split("_")[1])
            if file_hip_ind in hip_inds["HIP"].values:
                relevant_files.append(file)
        self.systems = []
        for system_file in tqdm(
            relevant_files, desc="Loading systems", position=0, leave=False
        ):
            if cache:
                cache_file = Path(cache_base, "exoverses", system_file.stem + ".p")
                if not cache_file.parent.exists():
                    cache_file.parent.mkdir(parents=True)
                if cache_file.exists():
                    with open(cache_file, "rb") as f:
                        system = dill.load(f)
                else:
                    system = ExovistaSystem(system_file, convert=convert, filter=filter)
                    with open(cache_file, "wb") as f:
                        dill.dump(system, f)
                self.systems.append(system)
            else:
                system = ExovistaSystem(system_file, convert=convert, filter=filter)
                if system is not None:
                    self.systems.append(system)

        # Get star ids
        self.ids = [system.star.id for system in self.systems]
        self.names = [system.star.name for system in self.systems]

        super().__init__()


def generate_systems(targetlist, path, seed, workers=12):
    settings = Settings.Settings(
        timemax=10.0, output_dir=path, seed=seed, npix=500, pixscale=0.0066
    )
    settings.emax = 0.1

    stars, nexozodis = load_stars.load_stars(targetlist, from_master=True)
    print("\n{0:d} stars in model ranges.".format(len(stars)))

    planets, albedos = generate_planets.generate_planets(
        stars, settings, force_earth=False
    )
    disks, compositions = generate_disks.generate_disks(
        stars, planets, settings, nexozodis=nexozodis
    )
    print("Generating scenes. (This may take a while.)")

    cores = min(workers, os.cpu_count())
    cores = min(cores, len(stars))
    percore = int(np.ceil(len(stars) / cores))
    if percore > 1 and percore * (cores - 1) >= len(stars):
        percore -= 1

    print(f"Using {cores} cores")
    pool = multiprocessing.Pool(cores)

    inputs = []
    for i in range(0, cores):
        imin = i * percore
        imax = (i + 1) * percore
        inputs.append(
            [
                stars.iloc[imin:imax],
                planets[imin:imax],
                disks[imin:imax],
                albedos[imin:imax],
                compositions[imin:imax],
                settings,
            ]
        )

    pool.starmap(generate_scene.generate_scene, [inputs[j][:] for j in range(0, cores)])
    pool.close()
    pool.join()


def runcmd(cmd, verbose=False):
    process = subprocess.Popen(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, shell=True
    )
    std_out, std_err = process.communicate()
    if verbose:
        print(std_out.strip(), std_err)
    pass


def get_data(universes=np.arange(1, 13), cache_location="data"):
    """
    This function gets all the exoVista data. It gets the csv file with the universe
    information and puts it in the "data/{universe_number}/target_database.csv".
    Then it goes through every url in the csv file and pulls that fits file into
    "data/{universe_number}/{file}.fits".
    """
    # Iterate over the different universes
    for n in tqdm(universes, position=0, desc="Universe", leave=False):
        universe_url = (
            "https://ckan-files.emac.gsfc.nasa.gov/"
            f"exovista/DEC21/{n}/target_database.csv"
        )
        Path(cache_location, str(n)).mkdir(parents=True, exist_ok=True)
        if not Path(cache_location, str(n), "target_database.csv").exists():
            runcmd(f"wget --directory-prefix=data/{n} {universe_url}", verbose=False)

        df = pd.read_csv(
            Path(cache_location, str(n), "target_database.csv"), low_memory=False
        )
        for i in tqdm(
            np.arange(1, df.shape[0]), position=1, desc="System", leave=False
        ):
            # Get file url
            fit_url = df.at[i, "URL"]

            # Create file path
            file_path = Path(cache_location, str(n), f"{fit_url.split('/')[-1]}")

            # If file doesn't exist then pull it
            if not file_path.exists():
                runcmd(
                    f"wget --directory-prefix=data/{n} {fit_url}",
                    verbose=False,
                )

            # Verify data and repull it if necessary
            pull_failure = True
            while pull_failure:
                pull_failure = check_data(file_path)
                if pull_failure:
                    runcmd(
                        f"wget --directory-prefix=data/{n} {fit_url}",
                        verbose=False,
                    )


def check_data(file):
    """
    This function verifies that all the fits files have the necessary data, sometimes
    they don't pull everything for some reason
    """
    system_info = fits.info(file, output=False)
    with open(file, "rb") as f:
        # read header of first extension
        h = fits.getheader(f, ext=0, memmap=False)
    n_ext = h["N_EXT"]  # get the largest extension
    failure = len(system_info) != n_ext + 1
    if failure:
        # Number of tables doesn't match the number of tables that the header
        # says exists, delete file
        file.unlink()
    return failure



import astropy.units as u
import numpy as np
from astropy.io import fits
from scipy.interpolate import interp1d

import exoverses.base as base


class ExovistaDisk(base.disk.Disk):
    def __init__(self, infile, fits_ext, star):
        self.star = star
        with open(infile, "rb") as f:
            obj_data, header = fits.getdata(f, ext=fits_ext, header=True, memmap=False)
            self._wavelengths = (
                fits.getdata(f, ext=fits_ext - 1, header=False, memmap=False) * u.um
            )

        self._lam_nm = self._wavelengths.to_value(u.nm)
        # The debris disk contrast cube (disk flux divided by star flux),
        # removing the last because it is a 2d map estimating the fractional
        # numerical noise in the contrast calculations
        self.contrast = obj_data[:-1]
        self.pixel_scale = header["PXSCLMAS"] * u.mas / u.pixel
        self.header = header

        self.disk_contrast_interp = interp1d(
            np.arange(len(self._lam_nm)),
            self.contrast,
            kind="cubic",
            axis=0,
        )

    def spec_flux_density(self, wavelengths, times):
        """
        Calculate the spectral flux density of the disk at the given times and
        wavelengths. Note that this is a per pixel calculation, and is
        inherenetly tied to the pixel scale of the disk.
        Args:
            wavelengths (astropy Quantity array):
                Wavelengths to calculate the spectral flux density at
            times (astropy Time array):
                Times to calculate the spectral flux density at
        Returns:
            disk_flux_density (astropy Quantity array):
                Spectral flux density of the disk at the given times in Jy. The
                shape of the cube is (Ntimes, Nwavelengths, Npix, Npix)
        """
        # Determine the indices the wavelengths we want to calculate the scene
        # are with respect to the disk wavelengths
        wavelengths_nm = wavelengths.to_value(u.nm)
        wavelengths_um = wavelengths.to_value(u.um)
        inds = np.searchsorted(self._lam_nm, wavelengths_nm) - 1

        # Determine the fractional index, with respect to the indices of the
        # wavelengths the disk was generated at, of the wavelengths we want to
        # calculate the scene at, since the disk interpolant is over the
        # wavelength index instead of the actual wavelength
        fracinds = inds + (
            np.log(wavelengths_um) - np.log(self._wavelengths[inds].to_value(u.um))
        ) / (
            np.log(self._wavelengths[inds + 1].to_value(u.um))
            - np.log(self._wavelengths[inds].to_value(u.um))
        )
        # Interpolate the disk to our desired wavelengths
        disk_contrast = self.disk_contrast_interp(fracinds).T

        # Instantiate the disk flux density cube
        shape = []
        single_eval = times.isscalar and wavelengths.isscalar

        if times.isscalar:
            shape.append(1)
        else:
            shape.append(len(times))
        if wavelengths.isscalar:
            shape.append(1)
        else:
            shape.append(len(wavelengths))
        shape.append(self.contrast.shape[1])
        shape.append(self.contrast.shape[2])
        shape = tuple(shape)

        disk_flux_density = np.zeros(shape) << u.Jy
        # Calculate the star's spectral flux density at the desired wavelengths
        star_flux_density = self.star.spec_flux_density(wavelengths, times)
        if not single_eval:
            for i, _ in enumerate(times):
                disk_flux_density[i] = np.multiply(
                    disk_contrast, star_flux_density[i]
                ).T
        else:
            disk_flux_density = np.multiply(disk_contrast, star_flux_density).T
        return disk_flux_density



__all__ = [
    "timetrans_to_timeperi",
    "timeperi_to_timetrans",
    "Msini",
    "rotate_vectors",
    "gen_rotate_to_sky_coords",
    "gen_rotate_to_local_ecliptic_coords",
    "drop_nan_coord_values",
    "add_units",
]

from .misc import (
    timetrans_to_timeperi,
    timeperi_to_timetrans,
    Msini,
    rotate_vectors,
    gen_rotate_to_sky_coords,
    gen_rotate_to_local_ecliptic_coords,
    drop_nan_coord_values,
    add_units,
)



import astropy.constants as const
import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.io.fits import getdata
from astropy.time import Time
from scipy.interpolate import RectBivariateSpline

import exoverses.base as base


class ExovistaStar(base.star.Star):
    """
    Class for the star in the exoVista systems
    """

    def __init__(self, infile):
        # Get the object's data from the fits file
        with open(infile, "rb") as f:
            obj_data, obj_header = getdata(f, ext=4, header=True, memmap=False)
            self._wavelengths = getdata(f, ext=0, header=False, memmap=False) * u.um

        # The times that the exovista scene was generated at, assuming
        # it starts at J2000
        self._t = Time(2000 + obj_data[:, 0], format="decimalyear")

        # Interpolate the x position of the star
        # self._x_pix = obj_data[:, 1] * u.pixel
        # self._y_pix = obj_data[:, 2] * u.pixel
        # _x_pix_interp = interp1d(self._t.jd, self._x_pix.value, kind="cubic")
        # _y_pix_interp = interp1d(self._t.jd, self._y_pix.value, kind="cubic")
        # self._x_pix_interp = lambda t: _x_pix_interp(t.jd)
        # self._y_pix_interp = lambda t: _y_pix_interp(t.jd)

        # Positions at times the exovista scene was generated at
        self._x = obj_data[:, 9] * u.AU
        self._y = obj_data[:, 10] * u.AU
        self._z = obj_data[:, 11] * u.AU

        # Velocities at times the exovista scene was generated at
        self._vx = obj_data[:, 12] * u.AU / u.yr
        self._vy = obj_data[:, 13] * u.AU / u.yr
        self._vz = obj_data[:, 14] * u.AU / u.yr

        # Load star's spectral flux density in Janskys
        self._star_flux_density = np.array(obj_data[:, 16:])
        self.star_flux_density_interp = RectBivariateSpline(
            self._wavelengths.to_value(u.nm),
            self._t.decimalyear,
            self._star_flux_density.T,
            kx=4,
            ky=4,
        )

        # System identifiers
        self.id = obj_header["ID"]
        self.name = f"HIP {obj_header['HIP']}"

        # System midplane information
        # I don't know why the PA is negative, but to match the exovista
        # output positions it has to be
        self.midplane_PA = obj_header["PA"] * u.deg  # Position angle
        self.midplane_I = obj_header["I"] * u.deg  # Inclination
        # if self.midplane_I < 0:
        #     breakpoint()

        # Proper motion
        self.PMRA = obj_header["PMRA"] * u.mas / u.yr
        self.PMDEC = obj_header["PMDEC"] * u.mas / u.yr

        # Celestial coordinates
        self.RA = obj_header["RA"] * u.deg
        self.DEC = obj_header["DEC"] * u.deg
        self.dist = obj_header["DIST"] * u.pc
        self.coords = SkyCoord(ra=self.RA, dec=self.DEC, distance=self.dist)

        # Spectral properties
        self.spectral_type = obj_header["TYPE"]
        self.MV = obj_header["M_V"] * u.mag  # Absolute V band mag

        # Commenting out for now, these are available but
        # not every star has all the information
        # self.Bmag = obj_header["BMAG"]
        # self.Vmag = obj_header["VMAG"]
        # self.Rmag = obj_header["RMAG"]
        # self.Imag = obj_header["IMAG"]
        # self.Jmag = obj_header["JMAG"]
        # self.Hmag = obj_header["HMAG"]
        # self.Kmag = obj_header["KMAG"]

        # Stellar properties
        # Bolometric luminosity
        self.luminosity = obj_header["LSTAR"] * u.Lsun
        self.effective_temperature = obj_header["TEFF"] * u.K
        self.angular_diameter = obj_header["ANGDIAM"] * u.mas
        self.mass = obj_header["MASS"] * u.M_sun
        self.radius = obj_header["RSTAR"] * u.R_sun
        self.logg = obj_header["LOGG"] * u.cm / u.s**2
        self.mu = self.mass * const.G
        self.pixel_scale = obj_header["PXSCLMAS"] * u.mas / u.pixel

        self.calc_jitter_terms()

    def spec_flux_density(self, wavelengths, times):
        """
        Calculate the spectral flux density of the star at the given wavelengths
        and times
        Args:
            wavelengths (astropy Quantity array):
                Wavelengths to calculate spectral flux density
            times (astropy Time array):
                Times to calculate spectral flux density
        Returns:
            F (astropy Quantity array):
                Spectral flux density values
        """
        return (
            self.star_flux_density_interp(
                wavelengths.to_value(u.nm), times.decimalyear
            ).T
            * u.Jy
        )



import numpy as np

from exoverses.base.system import System
from exoverses.exosims.planet import ExosimsPlanet
from exoverses.exosims.star import ExosimsStar


class ExosimsSystem(System):
    """
    Class for the whole stellar system
    """

    def __init__(self, SU, sInd, t0):
        # Create star object
        self.star = ExosimsStar(SU, sInd)
        self.planets = []
        self.pInds = np.where(SU.plan2star == sInd)[0]
        # loop over all planets
        for pInd in self.pInds:
            self.planets.append(ExosimsPlanet(SU, self.star, pInd, t0))

        self.cleanup()
        self.origin = "EXOSIMS"
        # Set up rebound simulation
        # self.sim = rebound.Simulation()
        # self.sim.G = const.G.value
        # self.sim.add(
        #     m=self.star.mass.decompose().value,
        #     x=self.star._x[0].decompose().value,
        #     y=self.star._y[0].decompose().value,
        #     z=self.star._z[0].decompose().value,
        #     vx=self.star._vx[0].decompose().value,
        #     vy=self.star._vy[0].decompose().value,
        #     vz=self.star._vz[0].decompose().value,
        # )
        # for planet in self.planets:
        #     self.sim.add(
        #         m=planet.mass.decompose().value,
        #         x=planet._x[0].decompose().value,
        #         y=planet._y[0].decompose().value,
        #         z=planet._z[0].decompose().value,
        #         vx=planet._vx[0].decompose().value,
        #         vy=planet._vy[0].decompose().value,
        #         vz=planet._vz[0].decompose().value,
        #     )
        # self.sim.move_to_com()



__all__ = [
    "ExosimsPlanet",
    "ExosimsStar",
    "ExosimsSystem",
    "ExosimsUniverse",
    "create_universe",
]

from .planet import ExosimsPlanet
from .star import ExosimsStar
from .system import ExosimsSystem
from .universe import ExosimsUniverse, create_universe



class Universe:
    """
    The base class for universe. Keeps track of the planetary systems.
    """

    def __init__(self) -> None:
        pass

    def __repr__(self):
        str = f"{self.type} universe\n"
        str += f"{len(self.systems)} systems loaded"
        return str



import json
from pathlib import Path

from astropy.time import Time
from EXOSIMS.util.get_module import get_module_from_specs
from tqdm import tqdm

from exoverses.base.universe import Universe
from exoverses.exosims.system import ExosimsSystem


def create_universe(universe_params):
    script_path = Path(universe_params["script"])
    with open(script_path) as f:
        specs = json.loads(f.read())
    assert "seed" in specs.keys(), (
        "For reproducibility the seed should not be randomized by EXOSIMS."
    )

    # Need to use SurveySimulation if we want to have a random seed
    SS = get_module_from_specs(specs, "SurveySimulation")(**specs)
    SU = SS.SimulatedUniverse
    universe_params["missionStart"] = specs["missionStart"]
    # SU = get_module_from_specs(specs, "SimulatedUniverse")(**specs)
    universe = ExosimsUniverse(SU, universe_params)
    return universe


class ExosimsUniverse(Universe):
    """
    Class for the whole EXOSIMS universe
    """

    def __init__(self, SU, params):
        """
        Args:
            path (str or Path):
                Location of all the system files. Should be something like "data/1/"
        """
        self.type = "EXOSIMS"
        self.SU = SU
        self.t0 = Time(params["missionStart"], format="mjd")
        # Load all systems
        sInds = SU.sInds
        if "nsystems" in params.keys():
            nsystems = params["nsystems"]
            sInds = sInds[:nsystems]
        else:
            sInds = SU.sInds
            nsystems = len(sInds)

        if "cache_path" in params.keys():
            self.cache_path = Path(params["cache_path"])

        self.systems = []
        for sInd in tqdm(sInds, desc="Loading systems", position=0, leave=False):
            system = ExosimsSystem(SU, sInd, self.t0)
            self.systems.append(system)

        # Get star ids
        self.ids = [system.star.id for system in self.systems]
        self.names = [system.star.name for system in self.systems]

        Universe.__init__(self)



__all__ = [
    "ExovistaDisk",
    "ExovistaPlanet",
    "ExovistaStar",
    "ExovistaSystem",
    "ExovistaUniverse",
    "create_universe",
]

from .disk import ExovistaDisk
from .planet import ExovistaPlanet
from .star import ExovistaStar
from .system import ExovistaSystem
from .universe import ExovistaUniverse, create_universe



import astropy.constants as const
import astropy.units as u
import numpy as np
import pandas as pd
from astropy.time import Time
from keplertools import fun as kt

import exoverses.util.misc as misc


class Planet:
    """
    Class for a planet
    """

    def __init__(self, planet_dict, star) -> None:
        for att, value in planet_dict.items():
            setattr(self, att, value)
        self.star = star
        self.solve_dependent_params()

    def __repr__(self):
        """
        Make dataframe with planet attributes
        """
        params = self.dump_params()
        res = {}
        for key, val in params.items():
            if isinstance(val, u.Quantity):
                res[key] = val.value
            elif isinstance(val, Time):
                res[key] = val.decimalyear
            else:
                res[key] = val

        # Create dataframe from res dictionary
        p_df = pd.DataFrame(res, index=[0])

        return f"{type(self).__name__} object\n{p_df}"

    def dump_params(self):
        params = {
            "t0": self.t0,
            "a": self.a,
            "e": self.e,
            "mass": self.mass,
            "radius": self.radius,
            "inc": self.inc,
            "W": self.W,
            "w": self.w,
            "M0": self.M0,
            "p": self.p,
        }
        return params

    def calc_vectors(
        self,
        t,
        return_r=True,
        return_v=False,
        coord_system="barycentric",
        convention="exovista",
    ):
        """
        Given a time, calculate the planet's barycentric position and/or
        velocity vectors
        Args:
            t (Time):
                Time to calculate the position vectors at
            return_r (bool):
                Whether to return the position vectors
            return_v (bool):
                Whether to return the velocity vectors
            coord_system (str):
                Coordinate system to return vectors in, either "barycentric" or
                "sky"
            convention (str):
                Orbital convention for rotations
        Returns:
            r(astropy Quantity array):
                3 x n stacked position vector in meters
            v(astroypy Quantity array):
                3 x n stacked velocity vector in m/s

        """
        # This will find the radial and velocity vectors at an epoch
        M = self.mean_anom(t)
        E = kt.eccanom(M.to(u.rad).value, self.e)
        orb_elem = (
            self.a.decompose().value,
            self.e,
            self.W.to(u.rad).value,
            self.inc.to(u.rad).value,
            self.w.to(u.rad).value,
        )
        a, e, Omega, inc, w = orb_elem
        if not np.isscalar(E):
            a = np.ones(len(E)) * a
            e = np.ones(len(E)) * e
            Omega = np.ones(len(E)) * Omega
            inc = np.ones(len(E)) * inc
            w = np.ones(len(E)) * w

        sinw = np.sin(w)
        cosw = np.cos(w)
        sinO = np.sin(Omega)
        cosO = np.cos(Omega)
        sininc = np.sin(inc)
        cosinc = np.cos(inc)
        asqrt1me2 = a * np.sqrt(1 - e**2)

        A = np.vstack(
            (
                a * (cosO * cosw - sinO * cosinc * sinw),
                a * (sinO * cosw + cosO * cosinc * sinw),
                a * sininc * sinw,
            )
        )

        B = np.vstack(
            (
                -asqrt1me2 * (cosO * sinw + sinO * cosinc * cosw),
                asqrt1me2 * (-sinO * sinw + cosO * cosinc * cosw),
                asqrt1me2 * sininc * cosw,
            )
        )

        # Calculate position vectors
        if return_r:
            if np.isscalar(self.mu) and not (np.isscalar(E)):
                r = np.matmul(A, np.array((np.cos(E) - e), ndmin=2)) + np.matmul(
                    B, np.array(np.sin(E), ndmin=2)
                )
            else:
                r = np.matmul(A, np.diag(np.cos(E) - e)) + np.matmul(
                    B, np.diag(np.sin(E))
                )
            if coord_system == "sky":
                r = misc.rotate_to_sky_coords(
                    r, self.star.midplane_I, self.star.midplane_PA, convention
                ).T
            r *= u.m

        # Calculate velocity vectors
        if return_v:
            if np.isscalar(self.mu) and not (np.isscalar(E)):
                v = (
                    np.matmul(-A, np.array(np.sin(E), ndmin=2))
                    + np.matmul(B, np.array(np.cos(E), ndmin=2))
                ) * np.tile(
                    np.sqrt(self.mu.decompose().value * a ** (-3.0))
                    / (1 - e * np.cos(E)),
                    (3, 1),
                )
            else:
                v = np.matmul(
                    np.matmul(-A, np.diag(np.sin(E)))
                    + np.matmul(B, np.diag(np.cos(E))),
                    np.diag(
                        np.sqrt(self.mu.decompose().value * a ** (-3.0))
                        / (1 - e * np.cos(E))
                    ),
                )
            if coord_system == "sky":
                v = misc.rotate_to_sky_coords(
                    v, self.star.midplane_I, self.star.midplane_PA, convention
                ).T

            v *= u.m / u.s

        if return_r and return_v:
            return r, v
        if return_r:
            return r
        if return_v:
            return v

    def rotate_to_sky_coords(self, vec):
        """
        Rotate the given vector to the sky coordinates, stub in the base class
        """
        return vec

    def mean_anom(self, times):
        """
        Calculate the mean anomaly at the given times
        Args:
            times (astropy Time array):
                Times to calculate mean anomaly

        Returns:
            M (astropy Quantity array):
                Planet's mean anomaly at t (radians)
        """
        M = ((self.n * ((times.jd - self.t0.jd) * u.d)).decompose() + self.M0) % (
            2 * np.pi * u.rad
        )
        return M

    def solve_dependent_params(self):
        self.mu = (const.G * (self.mass + self.star.mass)).decompose()
        self.T = (2 * np.pi * np.sqrt(self.a**3 / self.mu)).to(u.d)
        self.w_p = self.w
        self.w_s = (self.w + np.pi * u.rad) % (2 * np.pi * u.rad)
        self.secosw = np.sqrt(self.e) * np.cos(self.w)
        self.sesinw = np.sqrt(self.e) * np.sin(self.w)
        T_e = (self.T * self.M0 / (2 * np.pi * u.rad)).decompose()
        self.T_p = self.t0 - T_e

        # Calculate the time of conjunction
        self.T_c = Time(
            misc.timeperi_to_timetrans(
                self.T_p.jd, self.T.value, self.e, self.w_s.value
            ),
            format="jd",
        )
        self.K = (
            (2 * np.pi * const.G / self.T) ** (1 / 3.0)
            * (self.mass * np.sin(self.inc) / self.star.mass ** (2 / 3.0))
            * (1 - self.e**2) ** (-1 / 2)
        ).decompose()

        # Mean angular motion
        self.n = (np.sqrt(self.mu / self.a**3)).decompose() * u.rad

    def classify_planet(self):
        """
        This determines the Kopparapu bin of the planet This is adapted from
        the EXOSIMS SubtypeCompleteness method classifyPlanets so that EXOSIMS
        isn't a mandatory import
        """

        # # Calculate the luminosity of the star, assuming main-sequence
        Rp = self.radius.to("earthRad").value
        #
        # Find the stellar flux at the planet's location as a fraction of earth's
        earth_Lp = const.L_sun / (1 * (1 + (0.0167**2) / 2)) ** 2
        Lp = (
            (
                self.star.luminosity
                / (self.a.to("AU").value * (1 + (self.e**2) / 2)) ** 2
                / earth_Lp
            )
            .decompose()
            .value
        )

        # Find Planet Rp range
        Rp_bins = np.array([0, 0.5, 1.0, 1.75, 3.5, 6.0, 14.3, 11.2 * 4.6])
        Rp_types = [
            "Sub-Rocky",
            "Rocky",
            "Super-Earth",
            "Sub-Neptune",
            "Sub-Jovian",
            "Jovian",
            "Super-Jovian",
        ]
        L_bins = np.array(
            [
                [1000, 182, 1.0, 0.28, 0.0035, 5e-5],
                [1000, 182, 1.0, 0.28, 0.0035, 5e-5],
                [1000, 187, 1.12, 0.30, 0.0030, 5e-5],
                [1000, 188, 1.15, 0.32, 0.0030, 5e-5],
                [1000, 220, 1.65, 0.45, 0.0030, 5e-5],
                [1000, 220, 1.65, 0.40, 0.0025, 5e-5],
                [1000, 220, 1.68, 0.45, 0.0025, 5e-5],
                [1000, 220, 1.68, 0.45, 0.0025, 5e-5],
            ]
        )
        # Find the bin of the radius
        Rp_bin = np.digitize(Rp, Rp_bins) - 1
        Rp_bin = max(0, min(Rp_bin, len(Rp_types) - 1))
        Rp_type = Rp_types[Rp_bin]
        # TODO Fix this to give correct when at edge cases since technically
        # they're not straight lines

        # # index of planet temp. cold,warm,hot
        L_types = ["Very Hot", "Hot", "Warm", "Cold", "Very Cold"]
        specific_L_bins = L_bins[Rp_bin, :]
        L_bin = np.digitize(Lp, specific_L_bins) - 1
        L_bin = max(0, min(L_bin, len(L_types) - 1))
        L_type = L_types[L_bin]
        self.subtype = f"{L_type} {Rp_type.lower()}"

        # Determine if the planet is Earth-like
        # Reverse luminosity scaling
        a = self.a.to("AU").value / np.sqrt(self.star.luminosity.to("Lsun").value)

        lower_a = 0.95
        upper_a = 1.67

        lower_R = 0.8 / np.sqrt(a)
        upper_R = 1.4
        self.is_earth = (lower_a <= a < upper_a) and (
            lower_R <= self.radius.to("earthRad").value < upper_R
        )

        # Determine EXOSIMS phase function information
        Rp_phase_inds = [0, 2, 2, 7, 7, 4, 4]
        self.exosims_phiIndex = Rp_phase_inds[Rp_bin]
        # TODO Inform the geometric albedo from this



import astropy.units as u
import numpy as np
from astropy import constants as c
from scipy.optimize import root
from scipy.spatial.transform import Rotation as R

"""
Functions from radvel, should either rework or go back to calling radvel
but it wouldn't install and I don't want to keep debugging
"""


def timetrans_to_timeperi(tc, per, ecc, omega):
    """
    from radvel
    Convert Time of Transit to Time of Periastron Passage
    Args:
        tc (float): time of transit
        per (float): period [days]
        ecc (float): eccentricity
        omega (float): longitude of periastron (radians)
    Returns:
        float: time of periastron passage
    """
    try:
        if ecc >= 1:
            return tc
    except ValueError:
        pass

    f = np.pi / 2 - omega
    ee = 2 * np.arctan(
        np.tan(f / 2) * np.sqrt((1 - ecc) / (1 + ecc))
    )  # eccentric anomaly
    tp = tc - per / (2 * np.pi) * (ee - ecc * np.sin(ee))  # time of periastron

    return tp


def timeperi_to_timetrans(tp, per, ecc, omega, secondary=False):
    """
    from radvel
    Convert Time of Periastron to Time of Transit
    Args:
        tp (float): time of periastron
        per (float): period [days]
        ecc (float): eccentricity
        omega (float): argument of peri (radians)
        secondary (bool): calculate time of secondary eclipse instead
    Returns:
        float: time of inferior conjunction (time of transit if system is transiting)
    """
    try:
        if ecc >= 1:
            return tp
    except ValueError:
        pass

    if secondary:
        f = 3 * np.pi / 2 - omega  # true anomaly during secondary eclipse
        ee = 2 * np.arctan(
            np.tan(f / 2) * np.sqrt((1 - ecc) / (1 + ecc))
        )  # eccentric anomaly

        # ensure that ee is between 0 and 2*pi (always the eclipse AFTER tp)
        if isinstance(ee, np.float64):
            ee = ee + 2 * np.pi
        else:
            ee[ee < 0.0] = ee + 2 * np.pi
    else:
        f = np.pi / 2 - omega  # true anomaly during transit
        ee = 2 * np.arctan(
            np.tan(f / 2) * np.sqrt((1 - ecc) / (1 + ecc))
        )  # eccentric anomaly

    tc = tp + per / (2 * np.pi) * (ee - ecc * np.sin(ee))  # time of conjunction

    return tc


def Msini(K, P, Mstar, e, Msini_units="earth"):
    """Calculate Msini
    from radvel
    Calculate Msini for a given K, P, stellar mass, and e
    Args:
        K (float or array: Doppler semi-amplitude [m/s]
        P (float or array): Orbital period [days]
        Mstar (float or array): Mass of star [Msun]
        e (float or array): eccentricity
        Msini_units (Optional[str]): Units of Msini {'earth','jupiter'}
            default: 'earth'
    Returns:
        float or array: Msini [units = Msini_units]
    """

    # convert inputs to array so they work with units
    P = np.array(P)
    Mstar = np.array(Mstar)
    K = np.array(K)
    e = np.array(e)
    G = c.G.value  # added gravitational constant
    Mjup = c.M_jup.value  # added Jupiter's mass
    Msun = c.M_sun.value  # added sun's mass
    Mstar = Mstar * Msun
    Mstar = np.array(Mstar)

    P_year = (P * u.d).to(u.year).value
    P = (P * u.d).to(u.second).value

    # First assume that Mp << Mstar
    K_0 = 28.4329
    Msini = (
        K
        / K_0
        * np.sqrt(1.0 - e**2.0)
        * (Mstar / Msun) ** (2.0 / 3.0)
        * P_year ** (1 / 3.0)
    )

    # Use correct calculation if any elements are >10% of the stellar mass
    if (np.array(((Msini * u.Mjup).to(u.M_sun) / (Mstar / Msun)).value > 0.10)).any():
        a = K * (((2 * (np.pi) * G) / P) ** (-1 / 3.0)) * np.sqrt(1 - (e**2))
        Msini = []
        if isinstance(P, float):
            n_elements = 1
        else:
            assert (
                (type(K) is type(P))
                and (type(P) is type(Mstar))
                and (type(Mstar) is type(e))
            ), "All input data types must match."
            assert K.size == P.size == Mstar.size == e.size, (
                "All input arrays must have the same length."
            )
            n_elements = len(P)
        for i in range(n_elements):

            def func(x):
                try:
                    return x - a[i] * ((Mstar[i] + x) ** (2 / 3.0))
                except IndexError:
                    return x - a * ((Mstar + x) ** (2 / 3.0))

            sol = root(func, Mjup)
            Msini.append(sol.x[0])

        Msini = np.array(Msini)
        Msini = Msini / Mjup

    if Msini_units.lower() == "jupiter":
        pass
    elif Msini_units.lower() == "earth":
        Msini = (Msini * u.M_jup).to(u.M_earth).value
    else:
        raise Exception("Msini_units must be 'earth', or 'jupiter'")

    return Msini


def rotate_vectors(vectors, axis, angle):
    """
    Rotates a set of Nx3 vectors around a single axis by a single angle
    Args:
        vectors (np.array):
            Nx3 array of vectors
        axis (list):
            3-element array specifying rotation axis (e.g. [0,0,1]
            for z-axis)
        angle (u.Quantity):
            Angle to rotate vectors by
    """
    rot = R.from_rotvec(np.array(axis) * angle.to(u.rad).value)
    return rot.apply(vectors)


def gen_rotate_to_sky_coords(
    vector, inclination, position_angle, convention="exovista"
):
    """
    Rotate from barycentric coordinates to plane of the sky, this is set up
    to match the exovista data

    Args:
        vec (np.array):
            Nx3 array of [x,y,z] vectors in barycentric coordinates
        inclination (astropy Quantity):
            Inclination of the system
        position_angle (astropy Quantity):
            Position angle of the system
        convention (str):
            Convention that describes the transormation to be applied

    Returns:
        vec (np.array):
            Nx3 array of vectors rotated to sky coordinates
    """
    if convention == "exovista":
        # Rotate around x axis with midplane inclination
        vector = rotate_vectors(vector, [1, 0, 0], -inclination)
        # Rotate around z axis with midplane position angle
        vector = rotate_vectors(vector, [0, 0, 1], position_angle)
        # Flip around z axis
        vector[:, 2] = -vector[:, 2]
    else:
        raise Exception(
            "Convention must be either 'exovista' or 'radvel', got {}".format(
                convention
            )
        )
    return vector


def gen_rotate_to_local_ecliptic_coords(
    vector, inclination, position_angle, convention="exovista"
):
    """
    Rotate from plane-of-the-sky coordinates to local ecliptic, this is set up
    to match the exovista data

    Args:
        vec (np.array):
            Nx3 array of [x,y,z] vectors in barycentric coordinates
        inclination (astropy Quantity):
            Inclination of the system
        position_angle (astropy Quantity):
            Position angle of the system
        convention (str):
            Convention that describes the transormation to be applied

    Returns:
        vec (np.array):
            Nx3 array of vectors rotated to sky coordinates
    """
    if convention == "exovista":
        # Flip around z axis
        vector[:, 2] = -vector[:, 2]

        # Rotate around z axis with midplane position angle
        vector = rotate_vectors(vector, [0, 0, 1], -position_angle)

        # Rotate around x axis with midplane inclination
        vector = rotate_vectors(vector, [1, 0, 0], inclination)

    else:
        raise Exception("No other conventions currently supported")
    return vector


# def drop_nan_coords(ds):
#     """
#     Removes coordinates from an xarray dataset that contain only NaN values.

#     This function iterates over all coordinates in the provided xarray dataset,
#     checks if each coordinate contains only NaN values, and removes those coordinates.
#     The check is performed using the isnull() method combined with the all() aggregation
#     function. Coordinates that are fully NaN are dropped from the dataset.

#     Args:
#         dataset (xr.Dataset):
#             The input xarray dataset from which NaN-only coordinates will be removed.

#     Returns:
#         xr.Dataset:
#             A new xarray dataset with NaN-only coordinates removed.
#     """
#     coords_to_drop = []
#     for coord in ds.coords:
#         coord_vals = ds[coord].values
#         # Check if all values in the coordinate are NaN
#         if ds[coord].isnull().all():
#             coords_to_drop.append(coord)

#     # Drop identified coordinates
#     for coord in coords_to_drop:
#         ds = ds.drop_vars(coord)

#     return ds


def drop_nan_coord_values(ds):
    """
    Drops coordinate values and their associated data if the data across all
    variables for those coordinate values are NaN.

    This function iterates over each coordinate in the dataset, checks each
    value of the coordinate to determine if all associated data across all
    variables are NaN, and if so, removes those coordinate values and their
    associated data.

    Args:
        ds (xr.Dataset):
            Input xarray dataset.

    Returns:
        xr.Dataset:
            Dataset with NaN-only coordinate values and their data removed.
    """
    for coord in list(ds.coords):
        unique_vals = ds[coord].values
        for val in unique_vals:
            # Mask for the current coordinate value across all data variables
            mask = ds.isel({coord: ds[coord] == val}).notnull()

            # Check if all data for this coordinate value are NaN across all variables
            if not mask.to_array().any():
                # Drop the coordinate value across all variables
                ds = ds.where(ds[coord] != val, drop=True)

    return ds


def add_units(
    ds, new_unit, vars=["x", "y", "z"], distance=None, pixel_scale=None, star_pixel=None
):
    """
    Add units to a dataset by adding a new data variable with the
    desired unit conversion.

    Args:
        ds (xarray.Dataset):
            The original dataset.
        new_unit (astropy.units.Unit):
            The target unit for conversion.
        vars (list of str):
            List of variable names to convert.
        distance (astropy.units.Quantity):
            Distance to system.
        pixel_scale (astropy.units.Quantity):
            Pixel scale of the data in angle/pixel units
        star_pixel (astropy.Quantity array):
            The [x,y] pixel where the star is located under the assumption
            that the star is at the center of the image
    """
    for var in vars:
        # Ensure the variable is in the dataset
        assert var in ds, f"Variable {var} not found in dataset."
        var_data = ds[var].copy()
        base_unit = var_data.unit
        base_data = var_data.data * base_unit
        # Handle different unit conversions
        if new_unit.physical_type == "length":
            converted_data = base_data.to(new_unit)
        elif new_unit.physical_type == "angle":
            assert distance is not None, (
                "Distance to system not provided for angular conversion."
            )
            converted_data = (
                np.arctan(base_data.to(u.m).value / distance.to(u.m).value) * u.rad
            ).to(new_unit)
        elif new_unit == "pixel":
            assert (
                (distance is not None)
                and (pixel_scale is not None)
                and (star_pixel is not None)
            ), "Distance to system and pixel scale must be provided."
            angular_data = (
                np.arctan(base_data.to(u.m).value / distance.to(u.m).value) * u.rad
            )
            converted_data = (angular_data / pixel_scale).to(new_unit) + star_pixel

        # Update the dataset with the converted data
        new_name = f"{var}({new_unit})"
        var_data.data = converted_data.value
        var_data.attrs["unit"] = new_unit

        ds[new_name] = var_data

    return ds



from importlib.resources import files
from pathlib import Path

import astropy.constants as const
import astropy.units as u
import pandas as pd


class Star:
    """
    The star of a system
    """

    def __init__(self, star_dict):
        self.spectral_type = star_dict["spectral_type"]
        self.dist = star_dict["dist"]
        self.name = star_dict["name"]
        self.mass = star_dict["mass"]

    def __repr__(self):
        return f"{type(self).__name__} object\n{self.name}"

    def calc_jitter_terms(self):
        # Granulation noise scaling from Luhn 2020
        self.sigma_gran = (
            1
            * u.m
            / u.s
            * (self.luminosity / const.L_sun) ** 0.5
            * (self.mass / const.M_sun) ** -1
            * (self.effective_temperature / (5777 * u.K)) ** -0.5
        ).decompose()

        # Magnetic activity noise scaling relation from Gupta 2021
        # Load rhk values from ExEP target list
        rhk_df = pd.read_csv(
            files("exoverses").joinpath(Path("data", "mamajek_rhk.csv"))
        )
        if self.name in rhk_df["HIP"].values:
            logrhk = rhk_df.loc[rhk_df["HIP"] == self.name]["logR'HK"].item()
            logsigma_mag = 1.66 * logrhk + 8.39
            self.sigma_mag = 10**logsigma_mag * u.m / u.s
        else:
            self.sigma_mag = None



import astropy.constants as const
import astropy.units as u
import numpy as np
import pandas as pd
import rebound
import xarray as xr
from astropy.time import Time
from keplertools import fun as kt
from tqdm import tqdm

import exoverses.util.misc as misc


class System:
    """
    Class for a single system. Must have a star and a list of planets.
    """

    def __init__(self, star=None, planets=None, disk=None) -> None:
        self.star = star
        self.planets = planets
        self.disk = disk
        if self.planets is not None:
            self.planet_cleanup()
        if self.star is not None:
            self.star_cleanup()

        self.origin = "Base"
        self.nbody_frame = "bary"

    def __repr__(self):
        return (
            f"{self.star.name}\tdist:{self.star.dist}\t"
            f"Type:{self.star.spectral_type}\n\n"
            f"Planets:\n{self.get_p_df()}"
        )

    @property
    def name(self):
        return self.star.name

    @property
    def has_planets(self):
        return len(self.planets) > 0

    def planet_cleanup(self):
        self.pInds = np.arange(len(self.planets))
        # Sort the planets in the system by semi-major axis
        a_vals = [planet.a.value for planet in self.planets]
        self.planets = np.array(self.planets)[np.argsort(a_vals)].tolist()
        self.pInds = self.pInds[np.argsort(a_vals)]

    def star_cleanup(self):
        if hasattr(self.star, "midplane_I"):
            self.midplane_I = self.star.midplane_I
        else:
            self.midplane_I = 0 * u.rad

        if hasattr(self.star, "midplane_PA"):
            self.midplane_PA = self.star.midplane_PA
        else:
            self.midplane_PA = 0 * u.rad

    def getpattr(self, attr):
        # Return array of all planet's attribute value, e.g. all semi-major
        # axis values
        if isinstance(getattr(self.planets[0], attr), u.Quantity):
            return [getattr(planet, attr).value for planet in self.planets] * getattr(
                self.planets[0], attr
            ).unit
        else:
            return [getattr(planet, attr) for planet in self.planets]

    def get_p_df(self):
        if self.has_planets:
            patts = [
                "K",
                "T",
                "secosw",
                "sesinw",
                "T_c",
                "a",
                "e",
                "inc",
                "W",
                "w",
                "M0",
                "t0",
                "mass",
                "radius",
            ]
            p_df = pd.DataFrame()
            for att in patts:
                pattr = self.getpattr(att)
                if isinstance(pattr, u.Quantity):
                    p_df[att] = pattr.value
                else:
                    p_df[att] = pattr

            return p_df
        else:
            return "No planets in system."

    def propagate_rv(self, times):
        """
        Propagates system at all times given and returns the radial velocity

        """
        # Get unique time values, multiple instruments can be scheduled to
        # observe at the same time
        times = Time(np.unique(times.jd), format="jd")

        syst_M = []
        for planet in self.planets:
            syst_M.append(planet.mean_anom(times))
        Marr = np.stack(syst_M).value
        rv_vals = (
            -kt.calc_RV_from_M(
                Marr,
                self.getpattr("e"),
                self.getpattr("w").to(u.rad).value,
                self.getpattr("K").value,
            )
            * u.m
            / u.s
        ).T
        df = pd.DataFrame(rv_vals)
        df["rv"] = df.sum(axis=1)
        df["t"] = times

        return df

    def create_dataset(self, times):
        """
        Create an xarray Dataset for the system's motion
        """
        timesd64 = times.datetime64
        n_stars = 1
        n_planets = len(self.planets)
        state_vars = ["x", "y", "z", "vx", "vy", "vz"]
        coords = {
            "time": timesd64,
            "object": ["star", "planet", "disk", "fit"],
            "index": np.arange(max(n_stars, n_planets)),
            "ref_frame": ["bary", "helio", "bary-sky", "helio-sky"],
            "prop": ["kepler", "nbody"],
        }

        data_vars = {
            var: (
                ["time", "object", "index", "ref_frame", "prop"],
                np.nan * np.ones((len(timesd64), 4, max(n_stars, n_planets), 4, 2)),
            )
            for var in state_vars
        }

        ds = xr.Dataset(data_vars, coords=coords)

        # Add units information
        for var in state_vars:
            ds[var].attrs["unit"] = u.m if var in ["x", "y", "z"] else u.m / u.s
        return ds

    def prop_kepler(self, times, ds):
        """
        Calculate the barycentric position and velocity vectors of the planets
        in the system at the given times.
        """
        for i, planet in enumerate(self.planets):
            # Calculate the position and velocity vectors
            r, v = planet.calc_vectors(times, return_v=True)
            for j, coord in enumerate(["x", "y", "z"]):
                ds[coord].loc[times.datetime64, "planet", i, "bary", "kepler"] = (
                    r[j].to(u.m).value
                )
            for j, coord in enumerate(["vx", "vy", "vz"]):
                ds[coord].loc[times.datetime64, "planet", i, "bary", "kepler"] = (
                    v[j].to(u.m / u.s).value
                )
        return ds

    def prop_nbody(self, times, ds):
        """
        Calculate the barycentric position and velocity vectors of all bodies
        in the system at the given times using rebound.
        """
        # Set up rebound simulation
        sim = rebound.Simulation()
        sim.G = const.G.decompose().value
        t0 = self.star._t[0]
        times_sec = (times - t0).sec
        times64 = times.datetime64

        # Add the star and planets to the simulation, currently assuming no
        # binary systems
        sim = self.add_objects_to_rebound(sim)
        sim.move_to_com()

        n_stars = 1
        for time_sec, time64 in tqdm(
            zip(times_sec, times64),
            total=len(times),
            desc="n-body system propagation",
            delay=0.5,
        ):
            sim.integrate(time_sec)
            for j, p in enumerate(sim.particles):
                object = "star" if j < n_stars else "planet"
                index = j if j < n_stars else j - n_stars

                for coord, value in zip(["x", "y", "z"], [p.x, p.y, p.z]):
                    ds[coord].loc[time64, object, index, self.nbody_frame, "nbody"] = (
                        value
                    )
                for coord, value in zip(["vx", "vy", "vz"], [p.vx, p.vy, p.vz]):
                    ds[coord].loc[time64, object, index, self.nbody_frame, "nbody"] = (
                        value
                    )
        return ds

    def add_heliocentric_motion(self, ds, prop="kepler"):
        """
        Subtract the star's motion from the system's motion to get the
        heliocentric motion of the planets
        """
        bary_frame = "bary"
        helio_frame = "helio"
        star_bary_r = (
            ds[["x", "y", "z"]]
            .sel(object="star", index=0, ref_frame=bary_frame, prop=prop)
            .to_array()
            .data
        )
        star_bary_v = (
            ds[["vx", "vy", "vz"]]
            .sel(object="star", index=0, ref_frame=bary_frame, prop=prop)
            .to_array()
            .data
        )

        assert np.all(~np.isnan(star_bary_r)), (
            "Star position vector has NaN. Cannot compute heliocentric frame."
            " Probably needs N-body propagation."
        )

        for object in ["star", "planet"]:
            for i in range(len(ds.index)):
                # Rotate the position vectors
                object_bary_r = (
                    ds[["x", "y", "z"]]
                    .sel(object=object, index=i, ref_frame=bary_frame, prop=prop)
                    .to_array()
                    .data
                )
                object_helio_r = object_bary_r - star_bary_r
                ds["x"].loc[:, object, i, helio_frame, prop] = object_helio_r[0]
                ds["y"].loc[:, object, i, helio_frame, prop] = object_helio_r[1]
                ds["z"].loc[:, object, i, helio_frame, prop] = object_helio_r[2]

                # Rotate the velocity vectors
                object_bary_v = (
                    ds[["vx", "vy", "vz"]]
                    .sel(object=object, index=i, ref_frame=bary_frame, prop=prop)
                    .to_array()
                    .data
                )
                object_helio_v = object_bary_v - star_bary_v
                ds["vx"].loc[:, object, i, helio_frame, prop] = object_helio_v[0]
                ds["vy"].loc[:, object, i, helio_frame, prop] = object_helio_v[1]
                ds["vz"].loc[:, object, i, helio_frame, prop] = object_helio_v[2]
        return ds

    def propagate(
        self,
        times,
        ds=None,
        prop="kepler",
        ref_frame="bary",
        convention="exovista",
        clean=False,
    ):
        """
        Wrapper to handle the various propagation methods
        """
        scalar_time = times.isscalar
        if scalar_time:
            times = Time([times])

        if ds is None:
            ds = self.create_dataset(times)
        # Get the barycentric motion
        if prop == "kepler":
            ds = self.prop_kepler(times, ds)
        elif prop == "nbody":
            ds = self.prop_nbody(times, ds)

        # Convert to the desired frame
        ds = self.convert_to_frame(
            ds, ref_frame=ref_frame, convention=convention, prop=prop
        )

        if scalar_time:
            ds = ds.isel(time=0)

        if clean:
            # Check for coordinate values that have all nan values
            # and drop them to save memory space
            ds = misc.drop_nan_coord_values(ds)
        return ds

    def convert_to_frame(
        self, ds, ref_frame="bary", convention="exovista", prop="kepler"
    ):
        if prop == "nbody" and self.nbody_frame == "bary-sky":
            # Add the base frame, local ecliptic barycentric
            ds = self.rotate_to_local_ecliptic_coords(ds, ref_frame="bary", prop=prop)

        if ref_frame in ["helio", "helio-sky"]:
            # Add the heliocentric motion
            ds = self.add_heliocentric_motion(ds, prop=prop)

        if ref_frame in ["bary-sky", "helio-sky"]:
            ds = self.rotate_to_sky_coords(
                ds, ref_frame=ref_frame, convention=convention, prop=prop
            )

        return ds

    def add_objects_to_rebound(self, sim):
        """
        Method that adds the star and planets to the rebound simulation
        """
        sim.add(
            m=self.star.mass.decompose().value,
            x=self.star._x[0].decompose().value,
            y=self.star._y[0].decompose().value,
            z=self.star._z[0].decompose().value,
            vx=self.star._vx[0].decompose().value,
            vy=self.star._vy[0].decompose().value,
            vz=self.star._vz[0].decompose().value,
        )
        for planet in self.planets:
            sim.add(
                m=planet.mass.decompose().value,
                x=planet._x[0].decompose().value,
                y=planet._y[0].decompose().value,
                z=planet._z[0].decompose().value,
                vx=planet._vx[0].decompose().value,
                vy=planet._vy[0].decompose().value,
                vz=planet._vz[0].decompose().value,
            )
        return sim

    def rotate_to_sky_coords(
        self, ds, ref_frame="bary-sky", convention="exovista", prop="kepler"
    ):
        """
        Rotate the state vectors to the sky coordinates from barycentric
        coordinates
        """
        if ref_frame == "bary-sky":
            base_frame = "bary"
        elif ref_frame == "helio-sky":
            base_frame = "helio"
        for object in ["star", "planet"]:
            for i in range(len(ds.index)):
                # Rotate the position vectors
                base_r = (
                    ds[["x", "y", "z"]]
                    .sel(object=object, index=i, ref_frame=base_frame, prop=prop)
                    .to_array()
                    .data.T
                )
                sky_r = misc.gen_rotate_to_sky_coords(
                    base_r,
                    self.midplane_I,
                    self.midplane_PA,
                    convention=convention,
                ).T
                ds["x"].loc[:, object, i, ref_frame, prop] = sky_r[0]
                ds["y"].loc[:, object, i, ref_frame, prop] = sky_r[1]
                ds["z"].loc[:, object, i, ref_frame, prop] = sky_r[2]

                # Rotate the velocity vectors
                base_v = (
                    ds[["vx", "vy", "vz"]]
                    .sel(object=object, index=i, ref_frame=base_frame, prop=prop)
                    .to_array()
                    .data.T
                )
                sky_v = misc.gen_rotate_to_sky_coords(
                    base_v,
                    self.midplane_I,
                    self.midplane_PA,
                    convention=convention,
                ).T
                ds["vx"].loc[:, object, i, ref_frame, prop] = sky_v[0]
                ds["vy"].loc[:, object, i, ref_frame, prop] = sky_v[1]
                ds["vz"].loc[:, object, i, ref_frame, prop] = sky_v[2]
        return ds

    def rotate_to_local_ecliptic_coords(
        self, ds, ref_frame="bary", convention="exovista", prop="kepler"
    ):
        """
        Rotate the state vectors to the sky coordinates from barycentric
        coordinates
        """
        if ref_frame == "bary":
            base_frame = "bary-sky"
        elif ref_frame == "helio":
            base_frame = "helio-sky"
        for object in ["star", "planet"]:
            for i in range(len(ds.index)):
                # Rotate the position vectors
                base_r = (
                    ds[["x", "y", "z"]]
                    .sel(object=object, index=i, ref_frame=base_frame, prop=prop)
                    .to_array()
                    .data.T
                )
                sky_r = misc.gen_rotate_to_local_ecliptic_coords(
                    base_r,
                    self.midplane_I,
                    self.midplane_PA,
                    convention=convention,
                ).T
                ds["x"].loc[:, object, i, ref_frame, prop] = sky_r[0]
                ds["y"].loc[:, object, i, ref_frame, prop] = sky_r[1]
                ds["z"].loc[:, object, i, ref_frame, prop] = sky_r[2]

                # Rotate the velocity vectors
                base_v = (
                    ds[["vx", "vy", "vz"]]
                    .sel(object=object, index=i, ref_frame=base_frame, prop=prop)
                    .to_array()
                    .data.T
                )
                sky_v = misc.gen_rotate_to_local_ecliptic_coords(
                    base_v,
                    self.midplane_I,
                    self.midplane_PA,
                    convention=convention,
                ).T
                ds["vx"].loc[:, object, i, ref_frame, prop] = sky_v[0]
                ds["vy"].loc[:, object, i, ref_frame, prop] = sky_v[1]
                ds["vz"].loc[:, object, i, ref_frame, prop] = sky_v[2]
        return ds



import astropy.constants as const
import astropy.units as u
import pandas as pd

from exoverses.base.star import Star


class ExosimsStar(Star):
    """
    Class for the star in the EXOSIMS systems
    """

    def __init__(self, SU, sInd):
        # Get the object's data from the fits file

        TL = SU.TargetList
        # Time data
        self._t = 0 * u.yr

        # Position data
        self._x = 0 * u.AU
        self._y = 0 * u.AU
        self._z = 0 * u.AU

        # Velocity data
        self._vx = 0 * u.AU / u.yr
        self._vy = 0 * u.AU / u.yr
        self._vz = 0 * u.AU / u.yr

        # System identifiers
        self.id = sInd
        self.name = TL.Name[sInd]

        # System midplane information
        # self.midplane_PA = (obj_header["PA"] * u.deg).to(u.rad)  # Position angle
        # self.midplane_I = (obj_header["I"] * u.deg).to(u.rad)  # Inclination

        # Proper motion
        self.PMRA = TL.pmra[sInd] * u.mas / u.yr
        self.PMDEC = TL.pmdec[sInd] * u.mas / u.yr

        # Celestial coordinates
        self.coords = TL.coords[sInd]
        self.RA = self.coords.ra
        self.DEC = self.coords.dec
        self.dist = TL.dist[sInd]

        # Spectral properties
        self.spectral_type = TL.spectral_class[sInd]
        self.MV = TL.Vmag[sInd]  # Absolute V band mag

        # Commenting out for now, these are available but
        # not every star has all the information
        # self.Bmag = obj_header["BMAG"]
        # self.Vmag = obj_header["VMAG"]
        # self.Rmag = obj_header["RMAG"]
        # self.Imag = obj_header["IMAG"]
        # self.Jmag = obj_header["JMAG"]
        # self.Hmag = obj_header["HMAG"]
        # self.Kmag = obj_header["KMAG"]

        # Stellar properties
        self.Lstar = TL.L[sInd] * u.Lsun  # Bolometric luminosity
        # self.Teff = obj_header["TEFF"] * u.K  # Effective temperature
        # self.angdiam = obj_header["ANGDIAM"] * u.K  # Angular diameter
        self.mass = TL.MsTrue[sInd]
        self.mu = self.mass * const.G
        # self.radius = obj_header["RSTAR"] * u.R_sun

        # Propagation table
        self.vectors = pd.DataFrame(
            {
                "t": [self._t.decompose().value],
                "x": [self._x.decompose().value],
                "y": [self._y.decompose().value],
                "z": [self._z.decompose().value],
                "vx": [self._vx.decompose().value],
                "vy": [self._vy.decompose().value],
                "vz": [self._vz.decompose().value],
            }
        )
